;Edited Herb Johnson Dec 2020-Mar 2021
;A Floating Point Subroutine Package for the 1802 in DDJ #37.
;by Paul Wasserman, Union, NJ 07083
;based on PL/M code in book
;Microprocessor Programming for Computer Hobbyists"
; by Neil Graham, Tab books
;Dec 28 2020  - reassembly produces binary identical to original
;Jan 7 2021 - add references to Graham, binaries still identical
;   - change labels from disasembled to named refs
;Mar 4 2021 - move names to RAM area

;routine calls from Dobbs article (underscored label)

;fp_sub	equ	02E9H		;AC-OP -> AC (correction from #44 Dr Dobbs letter)
;fp_mul	equ	02B7H		;AC*OP -> AC
;fp_div	equ	034CH		;AC/OP -> AC
;clr_ac	equ	03C3H		;0 -> AC
;clr_op	equ	03D9H		;0 -> OP
;fp_add	equ	0400H		;AC+OP -> AC 
;dec_bin	equ	0500H		;D -> AC
;bin_dec	equ	0598H		;AC    -> D

;next three calls, 
;byte following call adr is variable number 0 to 42 decimal
;entered with R6 w/address of byte w/var number
;var_ac	equ	06DAH		;V     -> AC
;var_op	equ	0705H		;V     -> OP
;ac_var	equ	072CH		;AC    -> V

;next call, 6 bytes following call address is float constant
;R6 points to address of that constant
;fconst_	equ	0754H		;MMMMSE -> AC

;SQRT(AC) also uses V27H, V26H (42D, 43D)
;sqrt_	equ	0773H		;SQRT(AC) -> AC

;routines found by disassembly (no underscore in lable)

;acPLUSop	equ	012EH	; op_man + ac_man -> ac_man, with carry propogated
;acMINop	equ	0164H	;ac_man - op_man -> ac_man
;acINC	equ	0195H	;add 1 to ac_man, is this ever called?
;acNEG	equ	01B0H	;negate ac_man
;acX2		equ	01CBH	;shift left (*2) ac_man
;mulqX2	equ	01E2H	;shift left (*2) mul_q
;acDIV2	equ	01F9H	;shift right (/2) ac_man
;mulqDIV2	equ	0210H	;shift right (/2) mul_q  
;opDIV2	equ	0227H	;shift right (/2) op_man  
;opMULac	equ	0240H	;op_man * ac_man -> ac_man via mul_q  
;FR_DIV	equ	028DH	;fraction divide, AC / OP? OP / AC? quotient in mul_q
;acNORM	equ	0300H	;normalize AC 
;mqTOac	equ	0378H	;mulq -> ac_man, part of fp_div
;opNEG	equ	03A8H	;negate op_man 
;acMUL10	equ	04CEH	;multiply ac_man by 10 decimal (overwrite op_man)

; useful constants
;PI		DB  08H, 0EEH, 87H, 64H, 00H, 82H 	;from Dr Dobbs example

	incl	"1802reg.asm"
;for A18 assembler, defines R0, R1, etc. 

page		equ	0000H		;change this to relocate code and storage

;note on storage of AC, MUL_Q, other variables
;these must all be in the same memory page (01xx, 02xx, 03xx, etc.)
;because they are addressed with the same high-byte address value.
;in original code all at 0100H and above
;address order of these absolutely critical

		org	page+100H

ac_man0	db	0A0H	;equ	0100H
ac_man1	db	0A1H	;equ	0101H
ac_man2	db	0A2H	;equ	0102H
ac_man3	db	0A3H	;equ	0103H

op_man0	db	0H	;equ	0104H
op_man1	db	001H	;equ	0105H
op_man2	db	002H	;equ	0106H
op_man3	db	003H	;equ	0107H

;in external use the mantissa will be 24 signifigant bits
; from Graham, but these routines use all 32 bits. 

mul_q0	db	0	;equ	0108H
mul_q1	db	0	;equ	0109H
mul_q2	db	0	;equ	010AH		; 
mul_q3	db	0	;equ	010BH		;multiplier quotient

		db	0	;safety

ac_sign	db	0A5H	;equ	010DH		;accumulator
ac_exp0	db	0AEH	;equ	010EH		;128D bias, 0 = zero mantissa
ac_exp1	db	0AEH	;equ	010FH		; overflow


op_sign	db	005H ;equ	0110H	
op_exp	db	00EH ;equ	0111H		;operand

whois		db	0	;equ	0112H		;temp byte value
whatis	db	0	;equ	0113H		;temp byte value

dm_sign	db	05DH	;equ	0114H		;decimal register
dm_man0	db	0D1H	;equ	0115H
dm_man1	db	0D2H	;equ	0116H
dm_man2	db	0D3H	;equ	0117H
dm_man3	db	0D4H	;equ	0118H
dm_man4	db	0D5H	;equ	0119H
dm_man5	db	0D6H	;equ	011AH
dm_man6	db	0D7H	;equ	011BH

de_sign	db	05EH	;equ	011CH		;decimal exponent
de_exp0	db	0E1H	;equ	011DH
de_exp1	db	0E2H	;equ	011EH

		db	0, 0, 0, 0, 0, 0, 0, 0	;011F-0126
		db	0, 0, 0, 0, 0, 0, 0		;0127-012D 
 
; Start code segment at 012EH	

; should probably have some very-small constant for effective-zero test
; operationally: if exponents are 2**30 (1EH) or 30-bits different,
; add or sub is not signifigant
; overflow/underflow somewhat covered by extended AC range	

acPLUSop:	; op_man + ac_man -> ac_man, with carry propogated
		;Graham p110
	LDI	high ac_man0	;012E-F8 01
	PHI	RA		;0130-BA
	LDI	low ac_man0	;0131-F8 00
	PLO	RA		;0133-AA
	SEX	RA		;0134-EA
	LDX			;0135-F0
	PHI	RB		;0136-BB get ac_man0 -> RB.1
	LDI	low op_man0	;0137-F8 04
	PLO	RA		;0139-AA
	GHI	RB		;013A-9B
	ADD			;013B-F4
	PHI	RB		;013C-BB ac_man0+op_man0-> RB.1
	LDI	low ac_man0	;013D-F8 00
	PLO	RA		;013F-AA
	GHI	RB		;0140-9B
	STR	RA		;0141-5A RB.1 -> ac_man0
	LDI	low op_man1	;0142-F8 05
	PLO	RA		;0144-AA
	LDX			;0145-F0
	PHI	RB		;0146-BB op_man1 ->RB.1
	LDI	low ac_man1	;0147-F8 01
	PLO	RA		;0149-AA
	GHI	RB		;014A-9B
	ADC			;014B-74
	STR	RA		;014C-5A ac_man1 + op_man1 -> ac_man1
	LDI	low op_man2	;014D-F8 06
	PLO	RA		;014F-AA
	LDX			;0150-F0
	PHI	RB		;0151-BB op_man2 ->RB.1
	LDI	low ac_man2	;0152-F8 02
	PLO	RA		;0154-AA
	GHI	RB		;0155-9B
	ADC			;0156-74
	STR	RA		;0157-5A ac_man2 + op_man2 -> ac_man2
	LDI	low op_man3	;0158-F8 07
	PLO	RA		;015A-AA
	LDX			;015B-F0
	PHI	RB		;015C-BB
	LDI	low ac_man3	;015D-F8 03
	PLO	RA		;015F-AA
	GHI	RB		;0160-9B
	ADC			;0161-74
	STR	RA		;0162-5A ac_man3 + op_man3 -> ac_man3
	SEP	R5		;0163-D5 RET

acMINop:	;ac_man - op_man -> ac_man, with borrow
		;Graham p111
	LDI	001H		;0164-F8 01
	PHI	RA		;0166-BA
	LDI	low op_man0	;0167-F8 04
	PLO	RA		;0169-AA
	SEX	RA		;016A-EA
	LDX			;016B-F0
	PHI	RB		;016C-BB op_man0 -> RB.1
	LDI	low ac_man0	;016D-F8 00
	PLO	RA		;016F-AA
	GHI	RB		;0170-9B
	SD			;0171-F5 
	STR	RA		;0172-5A ac_man0 - op_man0 -> ac_man0
	LDI	low op_man1	;0173-F8 05
	PLO	RA		;0175-AA
	LDX			;0176-F0
	PHI	RB		;0177-BB op_man1 ->RB.1
	LDI	low ac_man1	;0178-F8 01
	PLO	RA		;017A-AA
	GHI	RB		;017B-9B
	SDB			;017C-75
	STR	RA		;017D-5A ac_man1 - op_man1 -> ac_man1
	LDI	low op_man2	;017E-F8 06
	PLO	RA		;0180-AA
	LDX			;0181-F0
	PHI	RB		;0182-BB op_man2 ->RB.1
	LDI	low ac_man2	;0183-F8 02
	PLO	RA		;0185-AA
	GHI	RB		;0186-9B
	SDB			;0187-75
	STR	RA		;0188-5A ac_man2 - op_man2 -> ac_man2
	LDI	low op_man3	;0189-F8 07
	PLO	RA		;018B-AA
	LDX			;018C-F0
	PHI	RB		;018D-BB  op_man3 ->RB.1
	LDI	low ac_man3	;018E-F8 03
	PLO	RA		;0190-AA
	GHI	RB		;0191-9B
	SDB			;0192-75
	STR	RA		;0193-5A ac_man3 - op_man3 -> ac_man3
	SEP	R5		;0194-D5 RET

acINC:	;add 1 to ac_man, is this ever called? Graham p112
	LDI	001H		;0195-F8 01
	PHI	RA		;0197-BA
	LDI	low ac_man0	;0198-F8 00
	PLO	RA		;019A-AA
	SEX	RA		;019B-EA
	LDI	001H		;019C-F8 01
	ADD			;019E-F4
	STR	RA		;019F-5A incr ac_man0 and set/clear carry
	INC	RA		;01A0-1A
	LDI	000H		;01A1-F8 00
	ADC			;01A3-74
	STR	RA		;01A4-5A carry into ac_man1
	INC	RA		;01A5-1A
	LDI	000H		;01A6-F8 00
	ADC			;01A8-74
	STR	RA		;01A9-5A carry into ac_man2
	INC	RA		;01AA-1A
	LDI	000H		;01AB-F8 00
	ADC			;01AD-74
	STR	RA		;01AE-5A carry into ac_man3
	SEP	R5		;01AF-D5 RET

acNEG:	;negate ac_man, Graham p112
	LDI	high ac_man0	;01B0-F8 01
	PHI	RA		;01B2-BA
	LDI	low ac_man0	;01B3-F8 00
	PLO	RA		;01B5-AA
	SEX	RA		;01B6-EA
	LDI	000H		;01B7-F8 00
	SM			;01B9-F7 
	STR	RA		;01BA-5A 0-ac_man0 -> ac_man0
	INC	RA		;01BB-1A ac_man1
	LDI	000H		;01BC-F8 00
	SMB			;01BE-77
	STR	RA		;01BF-5A 0-ac_man1 -> ac_man1
	INC	RA		;01C0-1A ac_man2
	LDI	000H		;01C1-F8 00
	SMB			;01C3-77
	STR	RA		;01C4-5A 0-ac_man2 -> ac_man2
	INC	RA		;01C5-1A ac_man3
	LDI	000H		;01C6-F8 00
	SMB			;01C8-77
	STR	RA		;01C9-5A  0-ac_man3 -> ac_man3
	SEP	R5		;01CA-D5 RET

acX2:		;shift left (*2) ac_man, Graham p 112-113
	LDI	001H		;01CB-F8 01
	PHI	RA		;01CD-BA
	LDI	low ac_man0	;01CE-F8 00
	PLO	RA		;01D0-AA
	SEX	RA		;01D1-EA
	LDX			;01D2-F0
	SHL			;01D3-FE 
	STR	RA		;01D4-5A shift left ac_man0
	INC	RA		;01D5-1A
	LDX			;01D6-F0
	SHLC			;01D7-7E
	STR	RA		;01D8-5A shift left ac_man1
	INC	RA		;01D9-1A
	LDX			;01DA-F0
	SHLC			;01DB-7E
	STR	RA		;01DC-5A shift left ac_man2
	INC	RA		;01DD-1A
	LDX			;01DE-F0
	SHLC			;01DF-7E
	STR	RA		;01E0-5A shift left ac_man3
	SEP	R5		;01E1-D5 RET

mulqX2:	;shift left (*2) mul_q, GRaham p112-113
	LDI	001H		;01E2-F8 01
	PHI	RA		;01E4-BA
	LDI	low mul_q0	;01E5-F8 08
	PLO	RA		;01E7-AA
	SEX	RA		;01E8-EA
	LDX			;01E9-F0
	SHL			;01EA-FE
	STR	RA		;01EB-5A shift left mul_q0
	INC	RA		;01EC-1A
	LDX			;01ED-F0
	SHLC			;01EE-7E
	STR	RA		;01EF-5A shift left mul_q1
	INC	RA		;01F0-1A
	LDX			;01F1-F0
	SHLC			;01F2-7E
	STR	RA		;01F3-5A shift left mul_q2
	INC	RA		;01F4-1A
	LDX			;01F5-F0
	SHLC			;01F6-7E
	STR	RA		;01F7-5A shift left mul_q3
	SEP	R5		;01F8-D5 RET

acDIV2:	;shift right (/2) ac_man, Graham p113-114
	LDI	001H		;01F9-F8 01
	PHI	RA		;01FB-BA
	LDI	low ac_man3	;01FC-F8 03
	PLO	RA		;01FE-AA
	SEX	RA		;01FF-EA
	LDX			;0200-F0
	SHR			;0201-F6
	STR	RA		;0202-5A shift right ac_man3
	DEC	RA		;0203-2A
	LDX			;0204-F0
	SHRC			;0205-76
	STR	RA		;0206-5A shift right ac_man2
	DEC	RA		;0207-2A
	LDX			;0208-F0
	SHRC			;0209-76
	STR	RA		;020A-5A shift right ac_man1
	DEC	RA		;020B-2A
	LDX			;020C-F0
	SHRC			;020D-76 shift right ac_man0
	STR	RA		;020E-5A
	SEP	R5		;020F-D5 RET

mulqDIV2:	;shift right (/2) mul_q , Graham p113-114
	LDI	high mul_q3	;0210-F8 01
	PHI	RA		;0212-BA
	LDI	low mul_q3	;0213-F8 0B
	PLO	RA		;0215-AA
	SEX	RA		;0216-EA
	LDX			;0217-F0 get mul_q3
	SHR			;0218-F6 div by 2
	STR	RA		;0219-5A save in mul_q3
	DEC	RA		;021A-2A RA = mul_q2
	LDX			;021B-F0
	SHRC			;021C-76 div by 2 with carry
	STR	RA		;021D-5A save in mul_q2
	DEC	RA		;021E-2A
	LDX			;021F-F0 get mul_q1
	SHRC			;0220-76 div by 2 with carry
	STR	RA		;0221-5A save in mul_q1
	DEC	RA		;0222-2A
	LDX			;0223-F0 get mul_q0
	SHRC			;0224-76 div by 2 with carry
	STR	RA		;0225-5A save in mul_q0
	SEP	R5		;0226-D5 RET

opDIV2:	; shift right (/2) op_man, Graham p113-114 
	LDI	high op_man3	;0227-F8 01
	PHI	RA		;0229-BA
	LDI	low op_man3	;022A-F8 07
	PLO	RA		;022C-AA
	SEX	RA		;022D-EA
	LDX			;022E-F0 get op_man3
	SHL			;022F-FE ?????
	LDX			;0230-F0
	SHRC			;0231-76 carry
	STR	RA		;0232-5A manipulates op_man3
	DEC	RA		;0233-2A
	LDX			;0234-F0 get op_man2
	SHRC			;0235-76 carry
	STR	RA		;0236-5A manipulates op_man2
	DEC	RA		;0237-2A
	LDX			;0238-F0 get op_man1
	SHRC			;0239-76 carry
	STR	RA		;023A-5A manipulates op_man1
	DEC	RA		;023B-2A
	LDX			;023C-F0 get op_man0
	SHRC			;023D-76 carry
	STR	RA		;023E-5A manipulates op_man0
	SEP	R5		;023F-D5 RET

opMULac:	;op_man * ac_man -> ac_man using mul_q 
		;Graham p114-115
	;move AC to MQ0
	LDI	high ac_man0	;0240-F8 01
	PHI	RA		;0242-BA
	LDI	low ac_man0	;0243-F8 00
	PLO	RA		;0245-AA
	SEX	RA		;0246-EA
	LDX			;0247-F0
	PHI	RB		;0248-BB ac_man0 ->RB.1
	LDI	low mul_q0	;0249-F8 08
	PLO	RA		;024B-AA
	GHI	RB		;024C-9B
	STR	RA		;024D-5A ac_man0 ->mul_q0
	LDI	low ac_man1	;024E-F8 01
	PLO	RA		;0250-AA
	LDX			;0251-F0
	PHI	RB		;0252-BB 
	LDI	low mul_q1	;0253-F8 09
	PLO	RA		;0255-AA
	GHI	RB		;0256-9B
	STR	RA		;0257-5A ac_man1 ->mul_q1
	LDI	low ac_man2	;0258-F8 02
	PLO	RA		;025A-AA
	LDX			;025B-F0
	PHI	RB		;025C-BB
	LDI	low mul_q2	;025D-F8 0A
	PLO	RA		;025F-AA
	GHI	R8		;0260-98
	STR	RA		;0261-5A ac_man2 ->mul_q2
	LDI	low ac_man3		;0262-F8 03
	PLO	RA		;0264-AA
	LDX			;0265-F0
	PHI	RB		;0266-BB
	LDI	low mul_q3	;0267-F8 0B
	PLO	RA		;0269-AA
	GHI	RB		;026A-9B
	STR	RA		;026B-5A ac_man3 ->mul_q3
	LDI	low ac_man0	;026C-F8 00
	PLO	RA		;026E-AA
	STR	RA		;026F-5A mul_q3 ->ac_man0
	INC	RA		;0270-1A
	STR	RA		;0271-5A mul_q3 ->ac_man1
	INC	RA		;0272-1A
	STR	RA		;0273-5A mul_q3 ->ac_man2
	INC	RA		;0274-1A
	STR	RA		;0275-5A mul_q3 ->ac_man3

;multiply with loop here on RB.0 for 31 bits
	LDI	001H		;0276-F8 01 ; init counter
	PLO	RB		;0278-AB
R0279	SEP	R4		;0279-D4 CALL+addr DW	01F9H
	DW	acDIV2 	;AC * 2
	SEP	R4		;D4 CALL+addr DW 0210H
	DW	mulqDIV2	;027D-02 10 shift right MQ
	BNF	R0284		;027F-3B 84 ;if bit shifted out of MQ is 1
	SEP	R4		;0281-D4 CALL+addr DW	012EH
	DW	acPLUSop	; add OP to AC 
R0284	GLO	RB		;0284-8B
	XRI	01FH		;0285-FB 1F ; 31 bits yet? 
	BZ	R028C		;0287-32 8C if so, return
	INC	RB		;0289-1B ;no, increment counter & loop
	BR	R0279		;028A-30 79
R028C	SEP	R5		;028C-D5 RET

FR_DIV:	; AC / OP fraction divide routine, quotient in mul_q
		;Graham p122 following
		;is there a test for OP == 0 or small?
	SEP	R4		;028D-D4 CALL+addr DW	0164H
	DW	acMINop	;call AC-OP ->AC
;loop here on RB.0
	LDI	001H		;0290-F8 01
	PLO	RB		;0292-AB ; init shift counter
R0293	SEP	R4		;0293-D4 CALL+addr DW	01E2H
	DW	mulqX2	; shift left MQ
	SEP	R4		;0296-D4 CALL+addr DW	01CBH
	DW	acX2		; shift left AC
	BNF	R02A0		;0299-3B A0 ;branch if no carry
	SEP	R4		;029B-D4 CALL 
	DW	acPLUSop	; else OP+AC->AC
	BR	R02AE		;029E-30 AE
	;if AC not negative
R02A0	SEP	R4		;02A0-D4 CALL+addr DW	0164H
	DW	acMINop	;AC-OP -> AC 
	LDI	high mul_q0	;02A3-F8 01
	PHI	RA		;02A5-BA
	LDI	low mul_q0	;02A6-F8 08
	PLO	RA		;02A8-AA
	SEX	RA		;02A9-EA
	LDX			;02AA-F0  get mul_q0
	ORI	001H		;02AB-F9 01 set LSB
	STR	RA		;02AD-5A  and store mul_q0
R02AE	GLO	RB		;02AE-8B
	XRI	020H		;02AF-FB 20 ; shifted 32D yet?
	BZ	R02B6		;02B1-32 B6 ; if so return
	INC	RB		;02B3-1B
	BR	R0293		;02B4-30 93 ;else incr and loop
R02B6	SEP	R5		;02B6-D5 RET

fp_mul:	;equ	02B7H		;AC*OP -> AC
	;float multiply, from Graham p137-139
     ;first set sign  as XOR of both signs
	LDI	high op_sign	;02B7-F8 01
	PHI	RA		;02B9-BA
	LDI	low op_sign	;02BA-F8 10
	PLO	RA		;02BC-AA
	SEX	RA		;02BD-EA
	LDX			;02BE-F0
	PHI	RB		;02BF-BB op_sign -> RB.1
	LDI	low ac_sign	;02C0-F8 0D
	PLO	RA		;02C2-AA
	GHI	RB		;02C3-9B
	XOR			;02C4-F3
	STR	RA		;02C5-5A op_sign XOR ac_sign -> ac_sign
	;sum exponents 
	LDI	low op_exp	;02C6-F8 11
	PLO	RA		;02C8-AA
	LDX			;02C9-F0
	PHI	RB		;02CA-BB op_exp -> RB.1
	LDI	low ac_exp0	;02CB-F8 0E
	PLO	RA		;02CD-AA
	GHI	RB		;02CE-9B
	ADD			;02CF-F4
	STR	RA		;02D0-5A op_exp + ac_exp0 -> ac_exp0
	INC	RA		;02D1-1A RA = ac_exp1
	LDX			;02D2-F0
	ADCI	000H		;02D3-7C 00
	STR	RA		;02D5-5A carry into ac_exp1
	LDI	low ac_exp0	;02D6-F8 0E
	PLO	RA		;02D8-AA
	LDX			;02D9-F0
	SMI	081H		;02DA-FF 81 ;minus 129 
	STR	RA		;02DC-5A
	INC	RA		;02DD-1A
	LDX			;02DE-F0
	SMBI	000H		;02DF-7F 00
	STR	RA		;02E1-5A
	SEP	R4		;02E2-D4 CALL+addr DW	0240H
	DW	opMULac		;multiply
	SEP	R4		;02E5-D4 CALL+addr DW	0300H
	DW	acNORM	;normalize
	SEP	R5		;02E8-D5 RET

fp_sub:	;equ	02E9H		;AC-OP -> AC
	;floating point subtract, from Graham p136
	;change the sign of OP and call float add
	LDI	high op_sign		;02E9-F8 01
	PHI	RA		;02EB-BA
	LDI	low op_sign	;02EC-F8 10
	PLO	RA		;02EE-AA
	SEX	RA		;02EF-EA
	LDX			;02F0-F0 read op_sign
	XRI	0FFH		;02F1-FB FF negate sign
	STR	RA		;02F3-5A and store in op_sign
	SEP	R4		;02F4-D4 CALL+addr DW	0400H
	DW	fp_add	; AC+OP -> AC to perform subtraction
	SEP	R5		;02F7-D5 RET
	; End code segment

	DB	0, 0, 0, 0, 0, 0, 0, 0	;02F8-02FF

acNORM:
	;normalize ac_man, exp, from Graham p136-138
     ;check for mantissa = 0 
	LDI	high ac_man0	;0300-F8 01
	PHI	RA		;0302-BA
	LDI	low ac_man0	;0303-F8 00
	PLO	RA		;0305-AA
	SEX	RA		;0306-EA
	LDX			;0307-F0 get ac_man0
	INC	RA		;0308-1A
	OR			;0309-F1 OR with ac_man1
	INC	RA		;030A-1A
	OR			;030B-F1 OR with ac_man2
	INC	RA		;030C-1A
	OR			;030D-F1 OR with ac_man3
	BNZ	R0319		;030E-3A 19 br if not zero
     ;if zero clear mantissa, set sign neg, return
	LDI	low ac_exp1	;0310-F8 0F  
	PLO	RA		;0312-AA 
	LDI	000H		;0313-F8 00
	STXD			;0315-73 00 ->ac_exp1
	STXD			;0316-73 00 ->ac_exp0
	STXD			;0317-73 00 ->ac_sign
	SEP	R5		;0318-D5 RET

; if bit 31 of mantissa not 0, then
;shft right and incr exponent
R0319	LDI	low ac_man3	;0319-F8 03
	PLO	RA		;031B-AA
	LDX			;031C-F0
	ANI	080H		;031D-FA 80 check MSBit of ac_man3
	BZ	R0331		;031F-32 31
	SEP	R4		;0321-D4 CALL+addr DW	01F9H
	DW	acDIV2	; shift right
	LDI	low ac_exp0	;0323-F8 0E 
	PLO	RA		;0326-AA
	SEX	RA		;0327-EA
	LDX			;0328-F0
	ADI	001H		;0329-FC 01
	STR	RA		;032B-5A increment ac_exp0
	INC	RA		;032C-1A
	LDX			;032D-F0
	ADCI	000H		;032E-7C 00
	STR	RA		;0330-5A and carry into ac_exp1
;else shft left (*2) until 1 in bit 30
;also decr exp for each shift
R0331	LDI	low ac_man3	;0331-F8 03
	PLO	RA		;0333-AA
	LDX			;0334-F0
	ANI	040H		;0335-FA 40 ; check bit 30
	BNZ	R034B		;0337-3A 4B
	SEP	R4		;0339-D4 CALL+addr DW	01CBH
	DW	acX2		;033A-01 CB ; AC_man *2
	LDI	low ac_exp0	;033c-F8 0E
	PLO	RA		;033E-AA
	SEX	RA		;033F-EA
	LDX			;0340-F0
	SMI	001H		;0341-FF 01
	STR	RA		;0343-5A decr ac_exp0
	INC	RA		;0344-1A
	LDX			;0345-F0
	SMBI	000H		;0346-7F 00
	STR	RA		;0348-5A and borrow from ac_exp1
	BR	R0331		;0349-30 31
R034B	SEP	R5		;034B-D5 RET

fp_div:	;equ	034CH		;AC/OP -> AC
	;float divide, from Graham p139-140
	LDI	high op_sign ;034C-F8 01
	PHI	RA		;034E-BA
	LDI	low op_sign ;034F-F8 10
	PLO	RA		;0351-AA
	SEX	RA		;0352-EA
	LDX			;0353-F0
	PHI	RB		;0354-BB op_sign ->RB.1
	LDI	low ac_sign ;0355-F8 0D
	PLO	RA		;0357-AA
	GHI	RB		;0358-9B
	XOR			;0359-F3 ac_sign XOR op_sign
	STR	RA		;035A-5A -> ac_sign
	LDI	low op_exp	;035B-F8 11
	PLO	RA		;035D-AA
	LDX			;035E-F0
	PHI	RB		;035F-BB op_exp -> RB.1
	LDI	low ac_exp0 ;0360-F8 0E
	PLO	RA		;0362-AA
	GHI	RB		;0363-9B
	SD			;0364-F5 ac_exp0-op_exp -> ac_exp0
	STR	RA		;0365-5A 
	INC	RA		;0366-1A point to ac_exp1
	LDX			;0367-F0
	SMBI	000H		;0368-7F 00 process borrow
	STR	RA		;036A-5A  and save in ac_exp1
	DEC	RA		;036B-2A point to ac_exp0
	LDX			;036C-F0
	ADI	080H		;036D-FC 80 ;add 128D
	STR	RA		;036F-5A
	INC	RA		;0370-1A point to ac_exp1
	LDX			;0371-F0
	ADCI	000H		;0372-7C 00 ;carry from add
	STR	RA		;0374-5A
	SEP	R4		;0375-D4 CALL+addr DW	028DH
	DW	FR_DIV	;0376-02 8D ;fraction divide
				; now must recover quotient from mq
				; then normalize after

mqTOac:	; this moves mulq -> ac_man & normalizes AC 
	LDI	high mul_q0 ;0378-F8 01
	PHI	RA		;037A-BA
	LDI	low mul_q0 ;037B-F8 08
	PLO	RA		;037D-AA
	SEX	RA		;037E-EA
	LDX			;037F-F0
	PHI	RB		;0380-BB mul_q0 -> RB.1
	LDI	low ac_man0 ;0381-F8 00
	PLO	RA		;0383-AA
	GHI	RB		;0384-9B 
	STR	RA		;0385-5A mul_q0 -> ac_man0
	LDI	low mul_q1	;0386-F8 09
	PLO	RA		;0388-AA
	LDX			;0389-F0
	PHI	RB		;038A-BB
	LDI	low ac_man1 ;038B-F8 01
	PLO	RA		;038D-AA
	GHI	RB		;038E-9B
	STR	RA		;038F-5A mul_q1 -> ac_man1
	LDI	low mul_q2 ;0390-F8 0A
	PLO	RA		;0392-AA
	LDX			;0393-F0
	PHI	RB		;0394-BB
	LDI	low ac_man2 ;0395-F8 02
	PLO	RA		;0397-AA
	GHI	RB		;0398-9B
	STR	RA		;0399-5A mul_q2 - > ac_man2
	LDI	low mul_q3	;039A-F8 0B
	PLO	RA		;039C-AA
	LDX			;039D-F0
	PHI	RB		;039E-BB
	LDI	low ac_man3 ;039F-F8 03
	PLO	RA		;03A1-AA
	GHI	RB		;03A2-9B
	STR	RA		;03A3-5A mul_q3 -> ac_man3
	SEP	R4		;03A4-D4 CALL+addr DW	0300H
	DW	acNORM		;03A5-03 00 ; normalize AC
	SEP	R5		;03A7-D5 RET

opNEG:	;negate op_man, as 0-op_man
	LDI	high op_man0	;03A8-F8 01
	PHI	RA		;03AA-BA
	LDI	low op_man0	;03AB-F8 04
	PLO	RA		;03AD-AA
	SEX	RA		;03AE-EA
	LDI	000H		;03AF-F8 00
	SM			;03B1-F7
	STR	RA		;03B2-5A (-) op_man0
	INC	RA		;03B3-1A
	LDI	000H		;03B4-F8 00
	SMB			;03B6-77
	STR	RA		;03B7-5A (-) op_man1
	INC	RA		;03B8-1A
	LDI	000H		;03B9-F8 00
	SMB			;03BB-77
	STR	RA		;03BC-5A (-) op_man2
	INC	RA		;03BD-1A
	LDI	000H		;03BE-F8 00
	SMB			;03C0-77
	STR	RA		;03C1-5A (-) op_man3
	SEP	R5		;03C2-D5 RET

clr_ac:	;equ	03C3H		;0 -> AC
	LDI	high ac_man3 ;03C3-F8 01
	PHI	RA		;03C5-BA
	LDI	low ac_man3 ;03C6-F8 03
	PLO	RA		;03C8-AA
	SEX	RA		;03C9-EA
	LDI	000H		;03CA-F8 00
	STXD			;03CC-73 clear ac_man3
	STXD			;03CD-73 clear ac_man2
	STXD			;03CE-73 clear ac_man1
	STR	RA		;03CF-5A clear ac_man0
	LDI	low ac_exp1 ;03D0-F8 0F
	PLO	RA		;03D2-AA
	LDI	000H		;03D3-F8 00
	STXD			;03D5-73 clear ac_exp1
	STXD			;03D6-73 clear ac_exp0
	STXD			;03D7-73 clear ac_sign
	SEP	R5		;03D8-D5 RET

clr_op:	;equ	03D9H		;0 -> OP
	LDI	high op_man3	;03D9-F8 01
	PHI	RA		;03DB-BA
	LDI	low op_man3	;03DC-F8 07
	PLO	RA		;03DE-AA
	SEX	RA		;03DF-EA
	LDI	000H		;03E0-F8 00
	STXD			;03E2-73 clear op_man3
	STXD			;03E3-73 op_man2
	STXD			;03E4-73 op_man1
	STXD			;03E5-73 op_man0
	LDI	low op_exp ;03E6-F8 11
	PLO	RA		;03E8-AA
	LDI	000H		;03E9-F8 00
	STXD			;03EB-73 clear op_exp
	STXD			;03EC-73 and op_sign
	SEP	R5		;03ED-D5 RET
; End code segment

	DB	0, 0, 0, 0, 0, 0, 0, 0	;03EE-03F5
	DB	0, 0, 0, 0, 0, 0, 0, 0	;03F6-03FD
	DB	0, 0					;03FE-03FF

fp_add:	;equ	0400H		;AC+OP -> AC  
	;floating point add, from Graham p132-136
	LDI	high op_exp	;0400-F8 01
	PHI	RA		;0402-BA
	LDI	low op_exp	;0403-F8 11
	PLO	RA		;0405-AA  RA -> op_exp
	SEX	RA		;0406-EA
	LDX			;0407-F0
	PHI	RB		;0408-BB op_exp -> RB
	LDI	low ac_exp0	;0409-F8 0E
	PLO	RA		;040B-AA 
	GHI	RB		;040C-9B
	SM			;040D-F7
	;compare exponents of AC and OP, largest exp will be in AC
	BNF	R0476		;040E-3B 76
	;if OPexp > ACexp, exchange AC and OP
	LDI	low op_sign	;0410-F8 10
	PLO	RA		;0412-AA  
	LDX			;0413-F0
	PHI	RB		;0414-BB op_sign -> RB.1
	LDI	low ac_sign	;0415-F8 0D
	PLO	RA		;0417-AA 
	LDX			;0418-F0
	PLO	RB		;0419-AB ac_sign -> RB.0
	GHI	RB		;041A-9B
	STR	RA		;041B-5A old op_sign -> ra_sign
	LDI	low op_sign	;041C-F8 10
	PLO	RA		;041E-AA 
	GLO	RB		;041F-8B
	STR	RA		;0420-5A old ra_sign -> op_sign
	;same code pattern for exp 
	LDI	low op_exp	;0421-F8 11
	PLO	RA		;0423-AA 
	LDX			;0424-F0
	PHI	RB		;0425-BB op_exp -> RB.1
	LDI	low ac_exp0	;0426-F8 0E
	PLO	RA		;0428-AA  
	LDX			;0429-F0
	PLO	RB		;042A-AB ac_exp0- >RB.0
	GHI	RB		;042B-9B 
	STR	RA		;042C-5A 
	LDI	low op_exp	;042D-F8 11
	PLO	RA		;042F-AA 
	GLO	RB		;0430-8B
	STR	RA		;0431-5A ;ac_exp0 <-> op_exp
	;ac_exp1 not preloaded, it fills from overflow from ac_exp0
	;same code pattern for mantissa
	LDI	low ac_man0	;0432-F8 00
	PLO	RA		;0434-AA 
	LDX			;0435-F0
	PHI	RB		;0436-BB ac_man0 ->RB.1
	LDI	low op_man0	;0437-F8 04
	PLO	RA		;0439-AA  
	LDX			;043A-F0
	PLO	RB		;043B-AB op_man0 ->RB.0
	GHI	RB		;043C-9B
	STR	RA		;043D-5A (AC man0) to (OP man0)
	LDI	low ac_man0	;043E-F8 00
	PLO	RA		;0440-AA  
	GLO	RB		;0441-8B
	STR	RA		;0442-5A (OP man0) to (AC man0)
	LDI	low ac_man1	;0443-F8 01
	PLO	RA		;0445- 
	LDX			;0446-F0  
	PHI	RB		;0447-BB ac_man1 -> RB.1
	LDI	low op_man1	;0448-F8 05
	PLO	RA		;044A-AA  
	LDX			;044B-F0 
	PLO	RB		;044C-AB op_man1 -> RB.0
	GHI	RB		;044D-9B
	STR	RA		;044E-5A (AC man1) to (OP man1)
	LDI	low ac_man1	;044F-F8 01
	PLO	RA		;0451-AA  
	GLO	RB		;0452-8B  
	STR	RA		;0453-5A (OP man1) to (AC man1)
	LDI	low ac_man2	;0454-F8 02
	PLO	RA		;0456-AA  
	LDX			;0457-F0
	PHI	RB		;0458-BB ac_man2 -> RB.1
	LDI	low op_man2	;0459-F8 06
	PLO	RA		;045B-AA 
	LDX			;045C-F0
	PLO	RB		;045D-AB op_man2 -> RB.0
	GHI	RB		;045E-9B
	STR	RA		;045F-5A (AC man2) to (OP man2)
	LDI	low ac_man2	;0460-F8 02
	PLO	RA		;0462-AA  
	GLO	RB		;0463-8B (OP man2) to (AC man2)
	STR	RA		;0464-5A
	LDI	low ac_man3	;0465-F8 03
	PLO	RA		;0467-AA  
	LDX			;0468-F0
	PHI	RB		;0469-BB ac_man3 ->RB.1
	LDI	low op_man3	;046A-F8 07
	PLO	RA		;046C-AA  
	LDX			;046D-F0
	PLO	RB		;046E-AB op_man3 ->RB.0
	GHI	RB		;046F-9B
	STR	RA		;0470-5A (AC man3) to (OP man3)
	LDI	low ac_man3	;0471-F8 03
	PLO	RA		;0473-AA  
	GLO	RB		;0474-8B
	STR	RA		;0475-5A (OP man3) to (AC man3)
	;compare larger ac_exp0 to op_exp, if 2**30 larger result already in ac
R0476	LDI	low ac_exp0	;0476-F8 0E
	PLO	RA		;0478-AA 
	LDX			;0479-F0
	PHI	RB		;047A-BB ac_exp0 -> RB.1
	LDI	low op_exp	;047B-F8 11
	PLO	RA		;047D-AA 
	GHI	RB		;047E-9B 
	SM			;047F-F7 ;compare ac_exp0 to op_exp
	PLO	RC		;0480-AC ;save difference in RC.0
	SMI	01EH		;0481-FF 1E
	BNF	R0486		;0483-3B 86
	;if exponents are 2**30 (1EH) different, add is not signifigant, exit done
	SEP	R5		;0485-D5 RET

	;check signs, if different negate the operand (Graham p135)
R0486	LDI	low ac_sign	;0486-F8 0D
	PLO	RA		;0488-AA  
	LDX			;0489-F0
	PHI	RB		;048A-BB ac_sign -> RB.1
	LDI	low op_sign	;048B-F8 10
	PLO	RA		;048D-AA  
	GHI	RB		;048E-9B
	XOR			;048F-F3 compare AC sign and OP sign
	BZ	R0495		;0490-32 95
	;compare signs, if same branch
	SEP	R4		;0492-D4 CALL+addr DW	03A8H
	DW	opNEG		; negate op_man

	;else decrement, call, and loop until exp equal
R0495	GLO	RC		;0495-8C diff of AC_exp0 and OP_exp
	BZ	R049E		;0496-32 9E
	DEC	RC		;0498-2C
	SEP	R4		;0499-D4 CALL+addr DW	0227H
	DW	opDIV2	;op_man / 2
	BR	R0495		;049C-30 95

R049E	SEP	R4		;049E-D4 CALL+addr DW	012EH
	DW	acPLUSop	;op_man + ac_man -> ac_man
	;if the signs of operands are different ...Graham p138
	LDI	high ac_sign	;04A1-F8 01
	PHI	RA		;04A3-BA
	LDI	low ac_sign	;04A4-F8 0D
	PLO	RA		;04A6-AA
	SEX	RA		;04A7-EA
	LDX			;04A8-F0
	PHI	RB		;04A9-BB ac_sign ->RB.1
	LDI	low op_sign	;04AA-F8 10
	PLO	RA		;04AC-AA
	GHI	RB		;04AD-9B
	XOR			;04AE-F3 compare ac_sign and op_sign
	BZ	R04CA		;04AF-32 CA
	LDI	low ac_man3	;04B1-F8 03
	PLO	RA		;04B3-AA
	LDX			;04B4-F0
	ANI	080H		;04B5-FA 80
	BZ	R04CA		;04B7-32 CA branch if not different

	;if signs different 
	SEP	R4		;04B9-D4 CALL+addr DW	01B0H
	DW	acNEG		; negate ac
	;and set ac_sign to op_sign
	LDI	high op_sign	;04BC-F8 01
	PHI	RA		;04BE-BA
	LDI	low op_sign	;04BF-F8 10
	PLO	RA		;04C1-AA
	SEX	RA		;04C2-EA
	LDX			;04C3-F0
	PHI	RB		;04C4-BB op_sign ->RB.1
	LDI	low ac_sign	;04C5-F8 0D
	PLO	RA		;04C7-AA
	GHI	RB		;04C8-9B
	STR	RA		;04C9-5A op_sign -> ac_sign

R04CA	SEP	R4		;04CA-D4 CALL+addr DW	0300H
	DW	acNORM		; normalize AC and end 
	SEP	R5		;04CD-D5 RET

acMUL10:	;multiply ac_man by 10 decimal (overwrite op_man)
		;Graham p 145 following
	SEP	R4		;04CE-D4 CALL+addr DW	01CBH
	DW	acX2		;shift left (*2) ac_man
	;move ac_man to op_man
	LDI	high ac_man0	;04D1-F8 01
	PHI	RA		;04D3-BA
	LDI	low ac_man0	;04D4-F8 00
	PLO	RA		;04D6-AA
	SEX	RA		;04D7-EA
	LDX			;04D8-F0
	PHI	RB		;04D9-BB ac_man0 -> RB.1
	INC	RA		;04DA-1A
	LDX			;04DB-F0
	PLO	RB		;04DC-AB ac_man1 -> RB.0
	LDI	low op_man0	;04DD-F8 04
	PLO	RA		;04DF-AA
	GHI	RB		;04E0-9B
	STR	RA		;04E1-5A (ac_man0) to (op_man0)
	INC	RA		;04E2-1A
	GLO	RB		;04E3-8B
	STR	RA		;04E4-5A (ac_man1) to (op_man1)
	LDI	low ac_man2	;04E5-F8 02
	PLO	RA		;04E7-AA
	LDX			;04E8-F0
	PHI	RB		;04E9-BB ac_man2 ->RB.1
	INC	RA		;04EA-1A
	LDX			;04EB-F0
	PLO	RB		;04EC-AB ac_man3 ->RB.0
	LDI	low op_man2	;04ED-F8 06
	PLO	RA		;04EF-AA
	GHI	RB		;04F0-9B
	STR	RA		;04F1-5A AC man2 -> OP man2)
	INC	RA		;04F2-1A
	GLO	RB		;04F3-8B
	STR	RA		;04F4-5A AC man3 -> OP man3
	SEP	R4		;04F5-D4 CALL+addr DW	01CBH
	DW	acX2		;shift left (*2) ac_man
	SEP	R4		;04F8
	DW	acX2		;shift left (*2) ac_man
	SEP	R4		;04FB
	DW	acPLUSop	;op_man + ac_man -> ac_man
	SEP	R5		;04FE-D5 RET

	DB	00		;04FF-00
	
dec_bin:	;equ	0500H		D -> AC routine, 
	;convert decimal to float from Graham p148 float input
	SEP	R4		;0500-D4 CALL+addr DW	03C3H
	DW	clr_ac	;clear AC 
	LDI	00H		;0503 F8 00
	PLO	RC		;0505-AC 0 -> RC.0
				;RC.0 index into Decimal register
	;read dm register as decimal BCD, add into AC
R0506	SEP	R4		;0506-D4 CALL+addr DW	03D9H
	DW	clr_op	;clear OP
	GLO	RC		;0509-8C get index
	ADI	low dm_man0; 050A-FC 15
	PLO	RA		;050C-AA
	LDI	high dm_man0 ;050D-F8 01
	PHI	RA		;050F-BA RA -> dm_man_n, from 0
	SEX	RA		;0510-EA
	LDX			;0511-F0
	ANI	00FH		;0512-FA 0F mask for BCD digit
	PHI	RB		;0514-BB low nybble -> RB.1
	LDI	low op_man0 ;0515-F8 04
	PLO	RA		;0517-AA RA -> op_man
	GHI	RB		;0518-9B
	STR	RA		;0519-5A low nybble -> op_man_n
	SEP	R4		;051A-D4 CALL+addr DW	012EH
	DW	acPLUSop	;op_man+ac_man -> ac_man
	SEP	R4		;051D-D4 CALL+addr DW	04CEH
	DW	acMUL10	;multiply ac_man by 10 decimal 
	INC	RC		;0520-1C increment index
	GLO	RC		;0521-8C
	XRI	07		;0522-FB 07 ;count through 7 digits
	BNZ	R0506		;0524-3A 06

	;AC now has binary value of Decimal register
	;must normalize AC to value below 1.0
	;initialize AC_exp = 2**31 
	LDI	high ac_exp0	;0526-F8 01
	PHI	RA		;0528-BA
	LDI	low ac_exp0	;0529-F8 0E
	PLO	RA		;052B-AA
	SEX	RA		;052C-EA
	LDI	09FH		;052D-F8 31 bits+128=159 
	STR	RA		;052F-5A 9FH -> AC exp0
	SEP	R4		;0530-D4 CALL+addr DW	0300H
	DW	acNORM	;normalize AC
	LDI	high dm_sign ;0533-F8 01
	PHI	RA		;0535-BA
	LDI	low dm_sign ;0536-F8 14
	PLO	RA		;0538-AA RA = dm_sign
	SEX	RA		;0539-EA
	LDX			;053A-F0 get DM sign 
	XRI	'+'		;053B-FB 2B
	BZ	R0545		;053D-32 45 if plus sign, branch
				;           and leave positive (0)
	LDI	low ac_sign	;053F-F8 0D
	PLO	RA		;0541-AA
	LDI	0FFH		;0542-F8 FF
	STR	RA		;0544-5A 0FFH -> AC sign, set minus
	;now convert decimal exponent
R0545	LDI	low de_exp0	;0545-F8 1D
	PLO	RA		;0547-AA RA = decimal exponent
	LDX			;0548-F0
	ANI	00FH		;0549-FA 0F mask to get 0-9 value
	SHL			;054B-FE value *2
	PHI	RB		;054C-BB de_exp0 masked -> RB.1
	LDI	low whatis	;054D-F8 13 point to whatis
	PLO	RA		;054F-AA
	GHI	RB		;0550-9B
	STR	RA		;0551-5A DE_exp0 -> whatis
	SHL			;0552-FE value * 4
	SHL			;0553-FE value * 8
	ADD			;0554-F4 (value*2)+ (value*8)
	STR	RA		;0555-5A value*10 -> whatis
	LDI	low de_exp1	;0556-F8 1E
	PLO	RA		;0558-AA
	LDX			;0559-F0
	ANI	00FH		;055A-FA 0F get next decimal exp
	PHI	RB		;055C-BB  value -> RB.1
	LDI	low whatis	;055D-F8 13
	PLO	RA		;055F-AA point to whatis
	GHI	RB		;0560-9B get previous value
	ADD			;0561-F4 and add
	PLO	RC		;0562-AC save converted exponent -> RC.0

	;load floating point 10 into OP as .1010B X 2**4
	SEP	R4		;0563-D4 CALL+addr DW	03D9H
	DW	clr_op	;clear OP
	LDI	high op_exp	;0566-F8 01
	PHI	RA		;0568-BA
	LDI	low op_exp	;0569-F8 11
	PLO	RA		;056B-AA  
	SEX	RA		;056C-EA
	LDI	084H		;056D-F8 84 set op_exp = +4  
	STR	RA		;056F-5A 84H -> op_exp
	LDI	low op_man3	;0570-F8 07
	PLO	RA		;0572-AA
	LDI	050H		;0573-F8 80D sets op_man .1010B
	STR	RA		;0575-5A 80D -> op_man3
	;divide AC by op(=10)
	SEP	R4		;0576-D4 CALL+addr DW	034CH
	DW	fp_div		;AC/OP -> AC
	LDI	high de_sign	;0579-F8 01
	PHI	RA		;057B-BA
	LDI	low de_sign	;057C-F8 1C
	PLO	RA		;057E-AA
	SEX	RA		;057F-EA
	LDX			;0580-F0
	XRI	'+'		;0581-FB 2B is it positive exponent
	BZ	R058E		;0583-32 8E branch if so

;if negative decimal exponent, divide by 10 repeatedly
R0585	GLO	RC		;0585-8C
	BZ	R0597		;0586-32 97
	DEC	RC		;0588-2C
	SEP	R4		;0589-D4 CALL+addr DW	034CH
	DW	fp_div		;058A-03 4C AC/OP -> AC
	BR	R0585		;058C-30 85
;if positive decimal exponent, multiply by 10 repeatedly
R058E	GLO	RC		;058E-8C
	BZ	R0597		;058F-32 97
	DEC	RC		;0591-2C
	SEP	R4		;0592-D4 CALL+addr DW	02B7H
	DW	fp_mul	;0593-02 B7 ;AC*OP -> AC
	BR	R058E		;0595-30 8E
R0597	SEP	R5		;0597-D5 RET

bin_dec:
	;AC->D routine, convert AC float into decimal number
	;from Graham p154 float output
	;
	;load OP with floating value 10 as .1010 X 2**4
	;load whois with 
	SEP	R4		;0598-D4 CALL+addr DW	03D9H
	DW	clr_op		;clear OP
	LDI	high op_exp	;059B-F8 01
	PHI	RA		;059D-BA
	LDI	low op_exp	;059E-F8 11
	PLO	RA		;05A0-AA
	SEX	RA		;05A1-EA
	LDI	084H		;05A2-F8 84 set op_exp = +4  
	STR	RA		;05A4-5A
	LDI	low op_man3	;05A5-F8 07
	PLO	RA		;05A7-AA
	LDI	050H		;05A8-F8 50  .1010 binary
	STR	RA		;05AA-5A 50H -> op_man3
	;put decimal 128+4 into whois to test exponent
	LDI	low whois		;05AB-F8 12
	PLO	RA		;05AD-AA
	LDI	07AH		;05AE-F8 7A = 132D = +4 exponent
	STR	RA		;05B0-5A 132D -> whois

	;keep /10D or *10D, based on exponent
R05B1	LDI	low ac_exp0	;05B1-F8 0E
	PLO	RA		;05B3-AA
	LDX			;05B4-F0 get whois value
	SMI	085H		;05B5-FF 85 test for 132D
	BNF	R05C9		;05B7-3B C9

	;if above, divide by 10 and incr whois
	SEP	R4		;05B9-D4 CALL+addr DW	034CH
	DW	fp_div	;AC/OP -> AC
	LDI	high whois	;05BC-F8 01
	PHI	RA		;05BE-BA
	LDI	low whois	;05BF-F8 12
	PLO	RA		;05C1-AA
	SEX	RA		;05C2-EA
	LDX			;05C3-F0
	ADI	01H		;05C4-FC 01
	STR	RA		;05C6-5A  incr whois
	BR	R05B1		;05C7-30 B1

	;if exponent below, multiply by 10 and decr whois 
R05C9	LDI	low ac_exp0	;05C9-F8 0E
	PLO	RA		;05CB-AA
	LDX			;05CC-F0
	SMI	08AH		;05CD-FF 8A ;test ac_exp0
	BDF	R05E1		;05CF-33 E1 ;branch if = 8AH
	SEP	R4		;05D1-D4 CALL+addr DW	02B7H
	DW	fp_mul	;AC*OP -> AC
	LDI	high whois	;05D4-F8 01
	PHI	RA		;05D6-BA
	LDI	low whois	;05D7-F8 12
	PLO	RA		;05D9-AA
	SEX	RA		;05DA-EA
	LDX			;05DB-F0
	SMI	00AH		;05DC-FF 0A
	STR	RA		;05DE-5A sub 10dec from whois
	BR	R05C9		;05DF-30 C9

	;ac_exp0 now almost normalized,
	;whois has expo adjustment
R05E1	LDI	low ac_exp0	;05E1-F8 0E
	PLO	RA		;05E3-AA
	LDX			;05E4-F0
	XRI	084H		;05E5-FB 84 is +4 exponent = +4?
	BNZ	R05FB		;05E7-3A FB branch if not

	;if = 84H, see if AC value too high, need one more division
	LDI	low ac_man3	;05E9-F8 03
	PLO	RA		;05EB-AA
	LDX			;05EC-F0
	SMI	051H		;05ED-FF 51 ;81 decimal, .1010 (10D)
	BNF	R05FB		;05EF-3B FB
	;if AC too high, incr whois and divide by op (10D) 
	LDI	low whois 	;05F1-F8 12
	PLO	RA		;05F3-AA
	LDX			;05F4-F0
	ADI	01		;05F5-FC 01
	STR	RA		;05F7-5A incr whois
	SEP	R4		;05F8-D4 CALL+addr DW	034CH
	DW	fp_div	;AC/OP -> AC

R05FB LDI	high ac_exp0	;05FB-F8 01
	PHI	RA		;05FD-BA
	LDI	low ac_exp0	;05FE-F8 0E
	PLO	RA		;0600-AA ; RA -> ac_exp0
	SEX	RA		;0601-EA
	LDX			;0602-F0
	; adjust AC depending on exponent

	XRI	084H		;0603-FB 84 is exp = +4  
	BNZ	R060C		;0605-3A 0C

	;if exponent = +4, adjust AC*2 
	SEP	R4		;0607-D4 CALL+addr DW	01CBH
	DW	acX2		;ac_man * 2
	BR	0621H		;060A-30 21

R060C	LDX			;060C-F0
	SMI	083H		;060D-FF 83 is ac_exp0 +3?
	BDF	R0621		;060F-33 21
	;if exponent < +3, adjust AC/2 and bump exponent
	SEP	R4		;0611-D4 CALL+addr DW	01F9H
	DW	acDIV2	;ac_man / 2
 	LDI	high ac_exp0	;0614-F8 01 
	PHI	RA		;0616-BA
	LDI	low ac_exp0	;0617-F8 0E
	PLO	RA		;0619-AA
	SEX	RA		;061A-EA
	LDX			;061B-F0 adjust exponent
	ADI	01		;061C-FC 01
	STR	RA		;061E-5A incr ac_exp0
	BR	R060C		;061F-30 0C

	;exponent resolved I think, 
	;round AC to 7 decimal digits by adding 
	;binary value .5 X 10**-6 or .10000101 X 
R0621	SEP	R4		;0621-D4 CALL+addr DW	03D9H
	DW	clr_op	;clear OP
	LDI	high op_man0	;0624-F8 01
	PHI	RA		;0626-BA
	LDI	low op_man0	;0627-F8 04
	PLO	RA		;0629-AA
	SEX	RA		;062A-EA
	LDI	086H		;062B-F8 86  .10000101
	STR	RA		;062D-5A op_man = binary value 
	SEP	R4		;062E-D4 CALL+addr DW	012EH
	DW	acPLUSop	;op_man+ac_man->ac_man w/carry

	;check to see if round up makes AC >= 10 decimal,
	;fix if it does 
	LDI	high ac_man3	;0631-F8 01
	PHI	RA		;0633-BA
	LDI	low ac_man3	;0634-F8 03
	PLO	RA		;0636-AA
	SEX	RA		;0637-EA
	LDX			;0638-F0
	SMI	0A0H		;0639-FF A0 is AC_man >=10 decimal?
	BNF	R0652		;063B-3B 52 nope

	;fix is to make AC = 1 decimal and bump exponent 
	LDI	low whois	;063D-F8 12
	PLO	RA		;063F-AA
	LDX			;0640-F0 get our exponent from whois
	ADI	01		;0641-FC 01
	STR	RA		;0643-5A incr whois
	LDI	low ac_man3	;0644-F8 03
	PLO	RA		;0646-AA point to ac_man3
	LDI	10H		;0647-F8 10
	STR	RA		;0649-5A 1 decimal -> ac_man3
	LDI	000H		;064A-F8 00
	DEC	RA		;064C-2A
	STR	RA		;064D-5A 00H -> ac_man2
	DEC	RA		;064E-2A
	STR	RA		;064F-5A 00H -> ac_man1
	DEC	RA		;0650-2A
	STR	RA		;0651-5A 00H -> ac_man0

	;now convert AC binary value (< 10) into decimal register
	;leftmost digit first
R0652	LDI	low ac_man3	;0652-F8 03
	PLO	RA		;0654-AA
	LDX			;0655-F0
	PHI	RB		;0656-BB ac_man3 ->RB.1
	LDI	low dm_man0	;0657-F8 15
	PLO	RA		;0659-AA point to dm_man0
	GHI	RB		;065A-9B ac_man3
	SHR			;065B-F6 shift out excess bits
	SHR			;065C-F6
	SHR			;065D-F6
	SHR			;065E-F6
	STR	RA		;065F-5A  and load into dm_man0

	;loop through rest of decimal digits
	;multiply AC_man by 10, grab 4 LSBs, make decimal digit
	LDI	06		;0660-F8 06 more decimal digits
	PLO	RC		;0662-AC 06 -> RC.0
R0663	GLO	RC		;0663-8C
	BZ	R0686		;0664-32 86  exit at zero (after dm_man0) 
	LDI	low ac_man3	;0666-F8 03
	PLO	RA		;0668-AA
	LDX			;0669-F0
	ANI	00FH		;066A-FA 0F
	STR	RA		;066C-5A mask off nybble (value 0-15)
	SEP	R4		;066D-D4 CALL+addr DW	04CEH
	DW	acMUL10	;;multiply ac_man * 10D 
	LDI	high ac_man3	;0670-F8 01
	PHI	RA		;0672-BA
	LDI	low ac_man3	;0673-F8 03
	PLO	RA		;0675-AA
	SEX	RA		;0676-EA
	LDX			;0677-F0
	PHI	RB		;0678-BB ac_man3 -> RB.1
	;compute DM address
	GLO	RC		;0679-8C - index 6 to 0 into DM
	SDI	low (dm_man6+1)	;067A-FD address of next dm_man byte
	PLO	RA		;067C-AA  
	GHI	RB		;067D-9B old ac_man3 value
	SHR			;067E-F6
	SHR			;067F-F6
	SHR			;0680-F6
	SHR			;0681-F6 move one nybble
	STR	RA		;0682-5A save in dm_man
	DEC	RC		;0683-2C count down index
	BR	R0663		;0684-30 63

	;AC_man made decimal, now sign
R0686	LDI	low ac_sign	;0686-F8 0D
	PLO	RA		;0688-AA
	LDX			;0689-F0 
	BZ	R0696		;068A-32 96 br if sign positive

	;ac sign negative, set DM sign to '-'
	LDI	02DH		;068C-F8 2D 
	PHI	RB		;068E-BB  ASCII '-' -> RB.1
	LDI	low dm_sign	;068F-F8 14
	PLO	RA		;0691-AA   
	GHI	RB		;0692-9B
	STR	RA		;0693-5A '-' -> dm_sign
	BR	R069E		;0694-30 9E

	;ac sign positive, set DM sign to '+'
R0696	LDI	02BH		;0696-F8 2B ASCII '+'
	PHI	RB		;0698-BB
	LDI	low dm_sign	;0699-F8 14
	PLO	RA		;069B-AA
	GHI	RB		;069C-9B
	STR	RA		;069D-5A ASCII '+' -> dm_sign

	;now set decimal exponent
R069E	LDI	000H		;069E-F8 00
	PLO	RC		;06A0-AC
	PLO	RB		;06A1-AB  00 ->RB.0 for later use
	LDI	low de_sign	;06A2-F8 1C
	PLO	RA		;06A4-AA
	LDI	02BH		;06A5-F8 2B ASCII '+'
	STR	RA		;06A7-5A  '+' ->de_sign
	LDI	low whois	;06A8-F8 12
	PLO	RA		;06AA-AA
	LDX			;06AB-F0 get the exponent again
	SMI	080H		;06AC-FF 80 is exponent negative?
	BDF	R06BD		;06AE-33 BD

	;exponent is negative, compliment and set de_sign '-'
	XRI	0FFH		;06B0-FB FF   
	ADI	01		;06B2-FC 01 2's complment
	STR	RA		;06B4-5A  save in whois 
	LDI	low de_sign	;06B5-F8 1C
	PLO	RA		;06B7-AA
	LDI	02DH		;06B8-F8 2D ASCII '-'
	STR	RA		;06BA-5A  .. into de_sign
	BR	R06BE		;06BB-30 BE

R06BD	STR	RA		;06BD-5A positive exponent into whois

	;now convert exponent. how many tens?
R06BE	LDI	low whois	;06BE-F8 12
	PLO	RA		;06C0-AA
R06C1	LDX			;06C1-F0
	SMI	00AH		;06C2-FF 0A whois - 10D
	BNF	R06CA		;06C4-3B CA
	STR	RA		;06C6-5A
	INC	RB		;06C7-1B count 10's in RB.0
	BR	R06C1		;06C8-30 C1

	;save 10's in de_exp0, remainder in de_exp1
R06CA	LDI	low de_exp0	;06CA-F8 1D
	PLO	RA		;06CC-AA
	GLO	RB		;06CD-8B
	STR	RA		;06CE-5A store tens in de_exp0
	LDI	low whois		;06CF-F8 12
	PLO	RA		;06D1-AA point to remainder in whois
	LDX			;06D2-F0
	PHI	RB		;06D3-BB whois -> RB.1
	LDI	low de_exp1	;06D4-F8 1E  
	PLO	RA		;06D6-AA
	GHI	RB		;06D7-9B
	STR	RA		;06D8-5A whois -> de_exp1
	SEP	R5		;06D9-D5 RET

var_ac:
	;V -> AC, V is variable 0 to 29H, R6 -> variable number index
	;format is MMMMSE, 4 mantissa, 1 sign, 1 exponent bytes
	;
	;determine address of Variable	
	LDI	high whois	;06DA-F8 01
	PHI	RA		;06DC-BA
	LDI	low whois	;06DD-F8 12
	PLO	RA		;06DF-AA RA -> temp byte location
	SEX	RA		;06E0-EA
	LDA	R6		;06E1-46 R6 has variable number index
	SHL			;06E2-FE times two
	STR	RA		;06E3-5A  
	SHL			;06E4-FE times four
	ADD			;06E5-F4 total is times six bytes
	PLO	RB		;06E6-AB RB.0 = variable offset addr
	LDI	high varbls	;06E7-F8 08
	PHI	RB		;06E9-BB RB = variable addr

	LDI	low ac_man0	;06EA-F8 00
	PLO	RA		;06EC-AA RA -> AC mantissa
	LDA	RB		;06ED-4B get var value and advance (incr) RB
	STR	RA		;06EE-5A var -> ac_man0
	INC	RA		;06EF-1A incr RA
	LDA	RB		;06F0-4B
	STR	RA		;06F1-5A var -> ac_man1
	INC	RA		;06F2-1A
	LDA	RB		;06F3-4B
	STR	RA		;06F4-5A var -> ac_man2
	INC	RA		;06F5-1A
	LDA	RB		;06F6-4B
	STR	RA		;06F7-5A var -> ac_man3
	LDI	low ac_sign	;06F8-F8 0D
	PLO	RA		;06FA-AA
	LDA	RB		;06FB-4B
	STR	RA		;06FC-5A var -> ac_sign
	INC	RA		;06FD-1A
	LDA	RB		;06FE-4B
	STR	RA		;06FF-5A var -> ac_exp0
	INC	RA		;0700-1A
	LDI	000H		;0701-F8 00
	STR	RA		;0703-5A  00 -> ac_exp1
	SEP	R5		;0704-D5 RET

var_op: 
	;V -> OP, V is variable 0 to 29H, R6 -> variable number
	;see code for V -> AC
	LDI	high whois	;0705-F8 01
	PHI	RA		;0707-BA
	LDI	low whois	;0708-F8 12
	PLO	RA		;070A-AA RA -> temp byte location
	SEX	RA		;070B-EA
	LDA	R6		;070C-46 get variable number
	SHL			;070D-FE
	STR	RA		;070E-5A
	SHL			;070F-FE
	ADD			;0710-F4
	PLO	RB		;0711-AB RB.0 = offset into varble array
	LDI	high varbls	;0712-F8 08
	PHI	RB		;0714-BB
	LDI	low op_man0	;0715-F8 04
	PLO	RA		;0717-AA RX -> varble
	LDA	RB		;0718-4B load byte, incr RB
	STR	RA		;0719-5A varble -> op_man0
	INC	RA		;071A-1A incr RA
	LDA	RB		;071B-4B
	STR	RA		;071C-5A varble -> op_man1
	INC	RA		;071D-1A
	LDA	RB		;071E-4B
	STR	RA		;071F-5A varble -> op_man2
	INC	RA		;0720-1A
	LDA	RB		;0721-4B
	STR	RA		;0722-5A varble -> op_man3
	LDI	low op_sign	;0723-F8 10
	PLO	RA		;0725-AA
	LDA	RB		;0726-4B
	STR	RA		;0727-5A  varble -> op_sign
	INC	RA		;0728-1A
	LDA	RB		;0729-4B
	STR	RA		;072A-5A  varble -> op_exp
	SEP	R5		;072B-D5 RET

ac_var:	;equ	072CH		;AC    -> V, R6 -> variable number
; store AC into variable 0-29H or 42D values
	LDI	high whois	;072C-F8 01
	PHI	RA		;072E-BA
	LDI	low whois	;072F-F8 12
	PLO	RA		;0731-AA RA -> temp byte  
	SEX	RA		;0732-EA
	LDA	R6		;0733-46
	SHL			;0734-FE
	STR	RA		;0735-5A
	SHL			;0736-FE
	ADD			;0737-F4 var numbr * 6 is index into array
	PLO	RB		;0738-AB variable index ->RB.0
	LDI	high varbls	;0739-F8 08
	PHI	RB		;073B-BB RB -> varble addr
	LDI	low ac_man0	;073C-F8 00
	PLO	RA		;073E-AA RX -> ac_man0
	LDA	RA		;073F-4A load byte, incr RA
	STR	RB		;0740-5B ac_man0 -> varble
	INC	RB		;0741-1B incr RB
	LDA	RA		;0742-4A
	STR	RB		;0743-5B ac_man1 -> varble
	INC	RB		;0744-1B
	LDA	RA		;0745-4A
	STR	RB		;0746-5B ac_man2 -> varble
	INC	RB		;0747-1B
	LDA	RA		;0748-4A
	STR	RB		;0749-5B ac_man3 -> varble
	INC	RB		;074A-1B
	LDI	low ac_sign	;074B-F8 0D
	PLO	RA		;074D-AA
	LDA	RA		;074E-4A
	STR	RB		;074F-5B ac_sign -> varble
	INC	RB		;0750-1B
	LDA	RA		;0751-4A
	STR	RB		;0752-5B ac_exp0 -> varble
	SEP	R5		;0753-D5 RET

fconst_:	;MMMMSE -> AC enter float value
		;R6 -> constant float value after call
	LDI	high ac_man0	;0754-F8 01
	PHI	RA		;0756-BA
	LDI	low ac_man0	;0757-F8 00
	PLO	RA		;0759-AA
	SEX	RA		;075A-EA
	LDA	R6		;075B-46
	STR	RA		;075C-5A const -> ac_man0
	INC	RA		;075D-1A
	LDA	R6		;075E-46
	STR	RA		;075F-5A const -> ac_man1
	INC	RA		;0760-1A
	LDA	R6		;0761-46
	STR	RA		;0762-5A const -> ac_man2
	INC	RA		;0763-1A
	LDA	R6		;0764-46
	STR	RA		;0765-5A const -> ac_man3
	LDI	low ac_sign	;0766-F8 0D
	PLO	RA		;0768-AA
	LDA	R6		;0769-46
	STR	RA		;076A-5A const -> ac_sign
	INC	RA		;076B-1A
	LDA	R6		;076C-46
	STR	RA		;076D-5A const -> ac_exp0
	INC	RA		;076E-1A
	LDI	000H		;076F-F8 00
	STR	RA		;0771-5A 00    -> ac_exp1
	SEP	R5		;0772-D5 RET

sqrt_:	;equ	0773H	;SQRT(AC) -> AC
	;Newton-Raphson iterative method, not in Graham
	;uses variables V26 V27 
	LDI	00AH		;0773-F8 0A
	PLO	RD		;0775-AD RD is counter from 10D
	SEP	R4		;0776-D4 CALL+addr DW	072CH
	DW	ac_var
	DB	26H		;0779-26 AC -> V26H or V38D
	LDI	high ac_exp0	;077A-F8 01
	PHI	RA		;077C-BA
	LDI	low ac_exp0	;077D-F8 0E
	PLO	RA		;077F-AA
	LDN	RA		;0780-0A
	SMI	01		;0781-FF 01
	STR	RA		;0783-5A divide AC by 2
	SEP	R4		;0784-D4 CALL+addr DW	072CH
	DW	ac_var
	DB	27H		;0787-27 AC ->V27H or V39D
R0788	GLO	RD		;0788-8D
	BZ	R07AE		;0789-32 AE exit at zero
	SEP	R4		;078B-D4 CALL+addr DW	06DAH
	DW	var_ac
	db	26H		;078E-26 V26H -> AC
	SEP	R4		;078F-D4 CALL+addr DW	0705H
	DW	var_op
	DB	27H		;0792-27 V27H -> OP
	SEP	R4		;0793-D4 CALL+addr DW	034CH
	DW	fp_div	;AC/OP -> AC
	SEP	R4		;0796-D4 CALL+addr DW	0705H
	DW	var_op
	DB	27H		;0799-27 V27H -> OP
	SEP	R4		;079A-D4 CALL+addr DW	0400H
	DW	fp_add	; AC+OP -> AC
	LDI	high ac_exp0	;079D-F8 01
	PHI	RA		;079F-BA
	LDI	low ac_exp0	;07A0-F8 0E
	PLO	RA		;07A2-AA
	LDN	RA		;07A3-0A
	SMI	01		;07A4-FF 01
	STR	RA		;07A6-5A decr ac_expo
	SEP	R4		;07A7-D4 CALL+addr DW	072CH
	DW	ac_var
	DB	27H		;07AA-27 AC -> V27H
	DEC	RD		;07AB-2D count down
	BR	R0788		;07AC-30 88
R07AE	SEP	R5		;07AE-D5 RET

	DB	0, 0		;07AF-00 00 
; End code segment

varbls	equ	page+0800H	;42 decimal 29 hex storage registers
					;of six bytes each, 252D bytes total.
					;so preserve data area to 08FFH. 
					;must all be in same memory page (00xx, 01xx, etc)
					;in original, were well above float routines at 0800H

	END

